#!/usr/bin/env python
#coding:utf-8
# Author:   htouvet
# Purpose: mount/unmount cifs home directories on pam script auth for X2GO
# Created: 17/04/2012

import sys,os,pwd
import logging
from optparse import OptionParser
import subprocess
import re

usage="""\
%prog action

TIS X2Go CIFS mounts / mounts cleanup for X2GO

action is either :
 mount : mount CIFS home directory for a user (default action, uses environment variables PAM_USER and PAM_AUTHTOK)
 cleanup : removed cifs mounts for users who have no x2go or ssh session
 x2gousers
 sshusers
 cifsmounts
 checknagios : check
 setup : setup pam_script to run mount on pam auth (pam_script_auth) and setup cronjob to cleanup cifs mounts
"""
version = "0.1"

try:
    defaultwg = re.search('search (.*)',open('/etc/resolv.conf','r').read()).groups()[0].replace('.local','').upper()
except:
    defaultwg = ''

parser = OptionParser(usage=usage,version="%prog " + version)
parser.add_option("-d","--dry-run",    dest="dry_run",    default=False, action='store_true', help="Dry run (default: %default)")
parser.add_option("-v","--verbose",    dest="verbose",    default=False, action='store_true', help="Verbose, print log on console (default: %default)")
parser.add_option("-u","--user",    dest="user",    default='', help="User to inspect (default: %default)")
parser.add_option("-l","--loglevel", dest="loglevel", default='info', type='choice',  choices=['debug','warning','info','error','critical'], metavar='LOGLEVEL',help="Loglevel (default: %default)")
parser.add_option("-w","--workgroup", dest="workgroup", default=defaultwg,help="Workgroup for CIFS mounts (default: %default)")
parser.add_option("-s","--cifsserver", dest="cifsserver", default = 'srvfichiers', help="File server for homes //srvfichiers/%user")
parser.add_option("-a","--addmount", dest="cifsmounts", action='append',default = [], help="Additional CIFS share //<cifserver>/<share> to mount on ~/<share>")
parser.add_option("-g","--logfile", dest="logfile", default='/var/log/%s.log' % os.path.basename(sys.argv[0]).replace('.py','') ,help="Log filename(default: %default)")

(options,args)=parser.parse_args()

# setup Logger
logger = logging.getLogger('tis-pammount')
# set loglevel
if options.loglevel in ('debug','warning','info','error','critical'):
    numeric_level = getattr(logging, options.loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % options.loglevel)
    logger.setLevel(numeric_level)

if options.logfile:
    hdlr = logging.FileHandler(options.logfile)
    hdlr.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(message)s'))
    logger.addHandler(hdlr) 

if options.verbose:
    hdlr = logging.StreamHandler() 
    hdlr.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(message)s'))
    logger.addHandler(hdlr)   


if len(args) > 1:
    print "ERROR : You must provide at most one action to perform"
    parser.print_usage()
    sys.exit(2)


#----------------------------------------------------------------------
def run(command):
    """run command and return a tuple (exitcode,stdout)"""
    logger.debug('running : %s' % command)
    p = subprocess.Popen("%s 2>&1" % command,stdout=subprocess.PIPE,shell=True)
    (stdout,stderr) = p.communicate()
    result = []
    if stderr <> None:
        result.append(stderr)
    if stdout <> None:
        result.append(stdout)
    logger.debug('  result : %i,%s' % (p.returncode,'\n'.join(result)))
    return (p.returncode,'\n'.join(result))


########################################################################
class X2gosession(object):
    """Parameter from running sessions"""
    agent_pid= ''
    session_id= ''
    display= ''
    server= ''
    status= ''
    init_time= ''
    cookie= ''
    client= ''
    gr_port= ''
    sound_port= ''
    last_time= ''
    uname= ''
    localtime= ''
    fs_port = ''

    #17095|jacardon-465-1334326413_stDGNOME_dp24|465|srvlts1.tranquilit.local|S|13.04.12 16:13:33|596a36be9ae8b2ad621155ca90f54756|82.245.60.114|31250|31251|16.04.12 22:51:51|jacardon|345139|31252
    def __init__(self,line):
        (self.agent_pid,
         self.session_id,
         self.display,
         self.server,
         self.status,
         self.init_time,
         self.cookie,
         self.client,
         self.gr_port,
         self.sound_port,
         self.last_time,
         self.uname,
         self.localtime,
         self.fs_port) = line.split('|')
    def __repr__(self):
        return repr(self.__dict__)

def x2sessions():
    return [ X2gosession(l) for l in run('x2golistsessions_root')[1].splitlines() if '|' in l ]

class Mount(object):
    mount = ''
    mountpoint = ''
    type = ''
    options = ''
    n1=0
    n2=0
    def __init__(self,st):
        """init from /proc/mounts line"""
        (self.mount,self.mountpoint,self.type,self.options,self.n1,self.n2) = st.split(' ')[0:6]
    def __repr__(self):
        return repr(self.__dict__)

def cifs_mounts():
    mounts = [ Mount(l) for l in run('cat /proc/mounts')[1].splitlines()]
    return [ m for m in mounts if m.type == 'cifs']

class Process(object):
    UID = ''
    PID = ''
    PPID = ''
    C = ''
    STIME = ''
    TTY = ''
    TIME = ''
    CMD = ''
    #----------------------------------------------------------------------
    #root       697  1186  0 18:19 ?        00:00:00 sshd: tisadmin [priv]
    def __init__(self,l):
        """from ps -f -C command"""
        (self.UID,self.PID,self.PPID,self.C,self.STIME,self.TTY,self.TIME) = l.split()[0:7]
        self.CMD = ' '.join(l.split()[7:])   
    def __repr__(self):
        return repr(self.__dict__)

def ssh_sessions():
    return [Process(l) for l in os.popen('ps -f -C sshd').read().splitlines()[1:]]

def cleanup_mounts(user=''):
    logger.debug('Cleanup CIFS mounts for %s' % (user or 'all'))
    users_x2go = set([s.uname for s in x2sessions() if s.status in ('R','S')])
    users_ssh = set([s.UID for s in ssh_sessions()])

    users = users_x2go.union(users_ssh)
    logger.debug('  current users : %s' % (users,))

    mounts = cifs_mounts()
    mounts.sort(key = lambda m : m.mountpoint,reverse=True)   
    logger.debug('  current mounts : %s' % (mounts,))

    to_unmount =[]
    to_unmount.extend(mounts)

    to_kill = []

    for m in mounts:
        assert(isinstance(m,Mount))
        parts = m.mountpoint.split('/')
        for p in parts:
            if p in users or (user and p <> user) :
                to_unmount.remove(m)
                if p in to_kill:
                    to_kill.remove(p)
                break
            elif not p in to_kill:
                try:
                    pw = pwd.getpwnam(p)
                    if pw and pw.pw_uid>=500:
                        to_kill.append(p)
                except:
                    pass

    if to_kill:
        for u in to_kill:
            cmd = 'pkill -u "%s"' % u
            if options.dry_run:
                print(cmd)
            else:
                (exitcode,result) = run(cmd)

    if not to_unmount:
        logger.debug("Nothing to unmount")
        print "Nothing to unmount"  

    for m in to_unmount:
        cmd = 'umount "%s"' % m.mountpoint
        logger.debug(cmd)
        if not options.dry_run:
            (exitcode,result) = run(cmd)
            logger.debug(result)
        else:
            print('umount "%s"' % m.mountpoint)

#----------------------------------------------------------------------
def cifsmount():
    """Execute the CIFS mount based on environment provided by pam_script auth"""
    pam_user = os.environ['PAM_USER']
    pam_authtok = os.environ['PAM_AUTHTOK']
    logger.debug('Starting CIFS home mount on auth for %s' % pam_user)
    myuid = pwd.getpwnam(pam_user).pw_uid
    fstab = open('/etc/fstab','r').read()
    if myuid >= 500:
        home = pwd.getpwnam(pam_user).pw_dir
        server = options.cifsserver
        workgroup = options.workgroup
        logger.info('Mounting home directories for %s on %s for workgroup %s' % (pam_user,server,workgroup))
        #cmd = 'grep -q "%(home)s " /proc/mounts || /sbin/mount.cifs //%(server)s/%(pam_user)s %(home)s -o "user=%(pam_user)s,workgroup=%(workgroup)s,password=%(pam_authtok)s"' % locals()
        mount = '//%(server)s/%(pam_user)s %(home)s cifs sec=krb5,users,noauto 0 0' % locals()
        if not mount in fstab:
            open('/etc/fstab','a').write('\n'+mount+'\n')
        cmd = 'grep -q "%(home)s " /proc/mounts || sudo -i -u %(pam_user)s mount %(home)s' % locals()
        logger.info(cmd)
        logger.info('%s' % os.environ)
        logger.info('%s' % os.listdir('/tmp'))

        if options.dry_run:
            print cmd
        else:
            (exitcode,result) = run(cmd)
            if result:
                logger.info(result)
            if exitcode:
                logger.error('Error %i' % exitcode)
        if options.cifsmounts:
            logger.info('Mounting additional directories for %s : %s' % (pam_user,options.cifsmounts))   
            for share in options.cifsmounts:
                mount = '//%(server)s/%(share)s %(home)s/%(share)s cifs sec=krb5,users,noauto 0 0' % locals()
                if not mount in fstab:
                    open('/etc/fstab','a').write('\n'+mount+'\n')
                cmd = 'grep -q "%(home)s/%(share)s " /proc/mounts || sudo -i -u %(pam_user)s mount %(home)s/%(share)s' % locals()
                if options.dry_run:
                    print cmd
                else:
                    (exitcode,result) = run(cmd)
                    if result:
                        logger.info(result)
                    if exitcode:
                        logger.error('Error %i' % exitcode)

        # success
        sys.exit(0)  

if __name__=='__main__':
    # options
    if args:
        action = args[0]  
    else:
        action = 'mount'
        
    if action == 'mount':
        cifsmount()
    elif action == 'cleanup':
        cleanup_mounts(user=options.user)
    elif action == 'x2gousers':
        sessions = x2sessions()
        for s in sessions:
            if not options.user or s.uname == options.user:
                print "%-20s %s %s" % (s.uname,s.client,s.status)
    elif action == 'sshusers':
        sessions = ssh_sessions()
        for s in sessions:
            if not options.user or s.UID == options.user:
                print "%-15s %-7s %s" % (s.UID,s.PID,s.CMD)
    elif action == 'cifsmounts':
        mounts = cifs_mounts()
        mounts.sort(key = lambda m : m.mountpoint,reverse=True)   

        for s in mounts:
            if not options.user or options.user in s.mountpoint:      
                print "%-50s %s" % (s.mountpoint,s.mount)

    else:
        print 'Unknown action %s' % action
        sys.exit(1)
